
# ----------------------------------------------------------------------
#  COMPONENT: xylegend - X/Y plot legend.
#
#  This widget is a legend for an X/Y plot, meant to view line graphs produced
#  as output from the run of a Rappture tool.  Use the "add" and
#  "delete" methods to control the curves showing on the plot.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2005  Purdue Research Foundation
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itk
package require BLT

option add *Xylegend.font \
    -*-helvetica-medium-r-normal-*-8-* widgetDefault

option add *Xylegend.Button.font \
    -*-helvetica-medium-r-normal-*-9-* widgetDefault

itcl::class ::Rappture::XyLegend {
    inherit itk::Widget

    private variable _autocolors {
        #0000cd
        #cd0000
        #00cd00
        #3a5fcd
        #cdcd00
        #cd1076
        #009acd
        #00c5cd
        #a2b5cd
        #7ac5cd
        #66cdaa
        #a2cd5a
        #cd9b9b
        #cdba96
        #cd3333
        #cd6600
        #cd8c95
        #cd00cd
        #9a32cd
        #6ca6cd
        #9ac0cd
        #9bcd9b
        #00cd66
        #cdc673
        #cdad00
        #cd5555
        #cd853f
        #cd7054
        #cd5b45
        #cd6889
        #cd69c9
        #551a8b
    }
    private variable _lastColorIndex ""
    private variable _dispatcher "" ;# dispatcher for !events
    private variable _graph	""
    private variable _tree	""
    private variable _diff	"";	# Polygon marker used for difference.
    private variable _rename	"";	# Node selected to be renamed.
    private variable _diffelements
    private variable _unmapHidden 0

    constructor {args} { graph }
    destructor {}

    public method reset {} 

    private method Add { elem label {flags ""}} 
    private method Average {} 
    private method BuildPopup { popup } 
    private method Check { menu }
    private method Delete { args } 
    private method Difference {} 
    private method Editor { option args }
    private method GetData { elem what } 
    private method Hide { args } 
    private method Lower { args } 
    private method Raise { args } 
    private method PopupMenu { x y }
    private method Recolor {} 
    private method Rename {} 
    private method SelectAll {}
    private method Show { args } 
    private method Toggle { args } 
    private method UnmapHidden {} 
}
                                                                                
itk::usual XyLegend {
    keep -background -foreground -cursor 
}

itk::usual ComboEntry {
}
itk::usual ComboMenu {
}
itk::usual TreeView {
    keep -foreground -cursor 
}
itk::usual Scrollset {
}
itk::usual ComboMenu {
}
itk::usual ComboEntry {
}

blt::bitmap define dot1 {
#define dot1_width 8
#define dot1_height 8
static unsigned char dot1_bits[] = {
   0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa};
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::XyLegend::constructor { graph args } {
    option add hull.width hull.height
    pack propagate $itk_component(hull) no
    itk_component add scrollbars {
        blt::scrollset $itk_interior.scrl \
            -yscrollbar $itk_interior.scrl.ys \
            -window $itk_interior.scrl.legend \
            -height 100
    }
    blt::tk::scrollbar $itk_interior.scrl.ys
    set _tree [blt::tree create]
    itk_component add legend {
        blt::treeview $itk_component(scrollbars).legend -linewidth 0 \
            -bg white -selectmode multiple \
            -highlightthickness 0 \
            -tree $_tree \
            -font "Arial 9" \
            -flat yes -separator / 
    }
    $itk_component(legend) column insert 0 "show" \
        -text "" -weight 0.0 -pad 0 -borderwidth 0
    $itk_component(legend) style checkbox "check" -showvalue no \
        -onvalue 0 -offvalue 1 
    $itk_component(legend) column configure "treeView" -justify left \
        -weight 1.0 -text "" -pad 0 -borderwidth 0 -edit no
    $itk_component(legend) column configure "show" -style "check" -pad {0 0} \
        -edit yes
    set m $itk_component(hull).mb.menu
    blt::combobutton $itk_component(hull).mb  \
        -icon [Rappture::icon wrench] \
        -menu $m -activerelief raised -relief flat

    itk_component add controls {
        blt::combomenu $m \
            -yscrollbar $m.ys \
            -xscrollbar $m.xs 
    }
    blt::tk::scrollbar $m.xs
    blt::tk::scrollbar $m.ys

    $m add -text "Hide" -command [itcl::code $this Hide] 
    $m add -text "Show" -command [itcl::code $this Show] 
    $m add -text "Toggle" -command [itcl::code $this Toggle] 
    $m add -text "Unmap Hidden" -command [itcl::code $this UnmapHidden] \
        -variable [itcl::scope _unmapHidden] \
        -type checkbutton
    $m add -type separator 
    $m add -text "Lower" -command [itcl::code $this Lower] 
    $m add -text "Raise" -command [itcl::code $this Raise] 
    $m add -type separator 
    $m add -text "Average" -command [itcl::code $this Average]
    $m add -text "Difference" -command [itcl::code $this Difference]
    $m add -type separator 
    $m add -text "Recolor" -command [itcl::code $this Recolor]
    $m add -text "Rename" -command [itcl::code $this Rename]
    $m add -text "Delete" -command [itcl::code $this Delete]
    set _unmapHidden [$graph cget -unmaphiddenelements]
    set _graph $graph
    set cmd [itcl::code $this Toggle current]
    $itk_component(legend) bind CheckBoxStyle <ButtonRelease-1> \
        [itcl::code [subst -nocommands {
            if { [%W edit -root -test %X %Y] } {
                %W edit -root %X %Y
                $this Toggle [%W nearest -root %X %Y]
                break
            }
        }]]
    bind $itk_component(legend) <Enter> { focus %W }
    $itk_component(legend) bind Entry <Control-KeyRelease-a> \
        [itcl::code $this SelectAll]
    $itk_component(legend) bind Entry <KeyRelease-Return> \
        +[itcl::code $this Toggle focus]
    $itk_component(legend) bind Entry <Escape> \
        "$itk_component(legend) selection clearall"
    $itk_component(legend) configure -selectcommand \
        [itcl::code $this Check $itk_component(controls)]
    bind $itk_component(legend) <ButtonRelease-3> \
        [itcl::code $this PopupMenu %X %Y]
    itk_component add editor {
        Rappture::Editor $itk_interior.editor \
            -activatecommand [itcl::code $this Editor activate] \
            -validatecommand [itcl::code $this Editor validate] \
            -applycommand [itcl::code $this Editor apply]
    }
    set _lastColorIndex [llength $_autocolors]
    blt::table $itk_interior \
        0,0 $itk_interior.mb -anchor w  \
        1,0 $itk_component(scrollbars) -fill both 
    blt::table configure $itk_interior r0 -resize none
    Check $itk_component(controls)
    eval itk_initialize $args
}

# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::XyLegend::destructor {} {
    foreach node [$_tree children root] {
        $_tree delete $node
    }
    if { $_diff != "" } {
        catch { $_graph marker delete $_diff }
    }
}

itcl::body Rappture::XyLegend::Add { elem label {flags ""} } {
    set hide [$_graph element cget $elem -hide]
    set im [image create picture]
    $_graph legend icon $elem $im
    set data(show) $hide
    set data(delete) [expr { $flags == "-delete" }]
    set node [$_tree insert root -at 0 -label $elem -data [array get data]]
    $itk_component(legend) entry configure $node -label $label -icon $im 
    update idletasks
    return $node
}

# ----------------------------------------------------------------------
# USAGE: reset <curve> ?<settings>?
#
# Clients use this to add a curve to the plot.  The optional <settings>
# are used to configure the plot.  Allowed settings are -color,
# -brightness, -width, -linestyle and -raise.
# ----------------------------------------------------------------------
itcl::body Rappture::XyLegend::reset {} {
    foreach node [$_tree children root] {
        $_tree delete $node
    }
    foreach elem [$_graph element show] {
        set label [$_graph element cget $elem -label]
        if { $label == "" } {
            set label $elem
        }
        Add $elem $label
    }
    $itk_component(legend) open -recurse root
    Check $itk_component(controls)
}

itcl::body Rappture::XyLegend::Hide { args } {
    if { $args == "" } {
        set nodes [$itk_component(legend) curselection]
    } else {
        set nodes $args
    }
    foreach node $nodes {
        set elem [$_tree label $node]
        if { ![$_graph element exists $elem] } {
            continue
        }
        $_graph element configure $elem -hide yes
        $_tree set $node "show" 1
    }
}

itcl::body Rappture::XyLegend::Show { args } {
    if { $args == "" } {
        set nodes [$itk_component(legend) curselection]
    } else {
        set nodes $args
    }
    foreach node $nodes {
        set elem [$_tree label $node]
        if { ![$_graph element exists $elem] } {
            continue
        }
        $_graph element configure $elem -hide no
        $_tree set $node "show" 0
    }
}

itcl::body Rappture::XyLegend::Toggle { args } {
    if { $args == "" } {
        set nodes [$itk_component(legend) curselection]
    } else {
        set nodes $args
    }
    foreach node $nodes {
        set elem [$_tree label $node]
        if { ![$_graph element exists $elem] } {
            continue
        }
        set hide [$_graph element cget $elem -hide]
        set hide [expr $hide==0]
        $_tree set $node "show" $hide
        $_graph element configure $elem -hide $hide
    }
}

itcl::body Rappture::XyLegend::Raise { args } {
    if { $args == "" } {
        set nodes [$itk_component(legend) curselection]
    } else {
        set nodes $args
    }
    set elements {}
    foreach node $nodes {
        set elem [$_tree label $node]
        set found($elem) 1
        set elements [linsert $elements 0 $elem]
    }
    foreach elem $elements {
        $_tree move [$_tree index $elem] 0 -at 0
    }
    set list {}
    foreach elem [$_graph element show] {
        if { [info exists found($elem)] }  {
            continue
        }
        lappend list $elem
    }
    $_graph element show [concat $list $elements]
}

itcl::body Rappture::XyLegend::Lower { args } {
    if { $args == "" } {
        set nodes [$itk_component(legend) curselection]
    } else {
        set nodes $args
    }
    set elements {}
    foreach node $nodes {
        set elem [$_tree label $node]
        set found($elem) 1
        set elements [linsert $elements 0 $elem]
    }
    set pos [$_tree degree 0]

    foreach elem $elements {
        incr pos -1
        $_tree move [$_tree index $elem] 0 -at $pos
    }

    set list {}
    foreach elem [$_graph element show] {
        if { [info exists found($elem)] }  {
            continue
        }
        lappend list $elem
    }
    $_graph element show [concat $elements $list]
}

itcl::body Rappture::XyLegend::Delete { args } {
    if { $args == "" } {
        set nodes [$itk_component(legend) curselection]
    } else {
        set nodes $args
    }
    set elements {}
    set delnodes {}
    foreach node $nodes {
        if { ![$_tree get $node "delete" 0] } {
            continue
        }
        set elem [$_tree label $node]
        lappend elements $elem
        lappend delnodes $node
        if { $_diff != "" && [info exists _diffelements($elem)] } {
            $_graph marker delete $_diff
            array unset _diffelements
            set _diff ""
        }
    }
    if { [llength $delnodes] > 0 } {
        eval $_tree delete $delnodes 
    }
    $itk_component(legend) selection clearall
    eval $_graph element delete $elements
}

itcl::body Rappture::XyLegend::Check { menu } {
    set nodes [$itk_component(legend) curselection]
    $menu item configure all -state disabled
    $menu item configure "Unmap Hidden" -state normal
    foreach node $nodes {
        if { [$_tree get $node "delete" 0] } {
            $menu item configure Delete -state normal
            break
        }
    }
    if { [$_tree degree 0] > 1  && [llength $nodes] > 0 } {
        foreach n { Raise Lower } {
            $menu item configure $n -state normal
        }
    }
    switch -- [llength $nodes] {
        0 {
        }
        1 {
            foreach n { Hide Show Toggle Rename Recolor } {
                $menu item configure $n -state normal
            }
        }
        2 {
            foreach n { Hide Show Toggle Difference Average Recolor } {
                $menu item configure $n -state normal
            }
        }
        default {
            foreach n { Hide Show Toggle Average Recolor } {
                $menu item configure $n -state normal
            }
        }
    }
}

itcl::body Rappture::XyLegend::GetData { elem axis } {
    set data [$_graph element cget $elem $axis]

    foreach { part1 part2} $data break
    if { [blt::vector names $part1] == $part1 } {
        return [$part1 values]
    } elseif { [blt::datatable names $part1] == $part1 } {
        return [$part1 column values $part2]
    }
    # Otherwise assume it's a list of numbers.
    return $data
}

itcl::body Rappture::XyLegend::Average {} {
    set nodes [$itk_component(legend) curselection]
    if { $nodes == "" } {
        return
    }
    set elements {}
    set sum [blt::vector create \#auto -command ""]

    set xcoords [blt::vector create \#auto -command ""]
    set ycoords [blt::vector create \#auto -command ""]

    blt::busy hold $itk_component(hull)
    update
    # Step 1. Get the x-values for each curve, then sort them to get the
    #	      unique values.

    set labels {}
    foreach node $nodes {
        set elem [$_tree label $node]
        set label [$_graph element cget $elem -label]
        $xcoords append [GetData $elem -x]
        set elements [linsert $elements 0 $elem]
        set labels [linsert $labels 0 $label]
    }
    # Sort the abscissas keeping unique values.
    $xcoords sort -uniq

    # Step 2. Now for each curve, generate a cubic spline of that curve
    #	      and interpolate to get the corresponding y-values for each 
    #	      abscissa.  Normally the abscissa are the same, so we're
    #	      interpolation the knots.

    set x [blt::vector create \#auto -command ""]
    set y [blt::vector create \#auto -command ""]
    $sum length [$xcoords length]

    foreach node $nodes {
        set elem [$_tree label $node]
        $x set [GetData $elem -x]
        $y set [GetData $elem -y]
        blt::spline natural $x $y $xcoords $ycoords

        # Sum the interpolated y-coordinate values.
        $sum expr "$sum + $ycoords" 
    }
    blt::vector destroy $x $y

    # Get the average
    $sum expr "$sum / [llength $elements]"

    # Step 3.  Create a new curve which is the average. Append it to the
    #	       the end.

    set count 0
    while {[$_graph element exists avg$count] } {
        incr count
    }
    set labels [lsort -dictionary $labels]
    set name "avg$count" 
    set label "Avg. [join $labels ,]"

    # Don't use the vector because we don't know when it will be cleaned up.

    if { $_lastColorIndex == 0 } {
        set _lastColorIndex [llength $_autocolors]
    }
    incr _lastColorIndex -1
    set color [lindex $_autocolors $_lastColorIndex]
    $_graph element create $name -label $label -x [$xcoords values]\
        -y [$sum values] -symbol scross -pixels 3 -color $color
    blt::vector destroy $xcoords $ycoords $sum
    set node [Add $name $label -delete]
    Raise $node
    blt::busy forget $itk_component(hull)
}

itcl::body Rappture::XyLegend::Difference {} {

    if { $_diff != "" } {
        $_graph marker delete $_diff
        set _diff ""
    }
    set nodes [$itk_component(legend) curselection]
    set elem1 [$_tree label [lindex $nodes 0]]
    set elem2 [$_tree label [lindex $nodes 1]]
    if { [info exists _diffelements($elem1)] && 
         [info exists _diffelements($elem2)] } {
        array unset _diffelements;	# Toggle the difference.
        return;				
    }
    array unset _diffelements
    set x [blt::vector create \#auto -command ""]
    set y [blt::vector create \#auto -command ""]
    set m [blt::vector create \#auto -command ""]

    $x append [GetData $elem1 -x]
    $y append [GetData $elem1 -y]
    $x sort -reverse $y 
    $x append [GetData $elem2 -x]
    $y append [GetData $elem2 -y]
    $m merge $x $y
    set _diff [$_graph marker create polygon \
                   -coords [$m values] \
                   -element $elem1 \
                   -stipple dot1 \
                   -outline "" -fill "#cd69c9"]
    blt::vector destroy $m $x $y
    set _diffelements($elem1) 1
    set _diffelements($elem2) 1
}


itcl::body Rappture::XyLegend::Recolor {} {
    set nodes [$itk_component(legend) curselection]
    if { $nodes == "" } {
        return
    }
    foreach node $nodes {
        set elem [$_tree label $node]
        if { $_lastColorIndex == 0 } {
            set _lastColorIndex [llength $_autocolors]
        }
        incr _lastColorIndex -1
        set color [lindex $_autocolors $_lastColorIndex]
        $_graph element configure $elem -color $color
        set im [$itk_component(legend) entry cget $node -icon]
        $_graph legend icon $elem $im
    }
}

itcl::body Rappture::XyLegend::UnmapHidden {} {
    $_graph configure -unmaphiddenelements $_unmapHidden
}

itcl::body Rappture::XyLegend::SelectAll { } {
    foreach node [$_tree children 0] {
        $itk_component(legend) selection set $node
    }  
}

itcl::body Rappture::XyLegend::Rename {} {
    Editor popup
}

# ----------------------------------------------------------------------
# USAGE: Editor popup
# USAGE: Editor activate
# USAGE: Editor validate <value>
# USAGE: Editor apply <value>
# USAGE: Editor menu <rootx> <rooty>
#
# Used internally to handle the various functions of the pop-up
# editor for the value of this gauge.
# ----------------------------------------------------------------------
itcl::body Rappture::XyLegend::Editor {option args} {
    switch -- $option {
        popup {
            $itk_component(editor) activate
        }
        activate {
            set _rename [$itk_component(legend) curselection]
            if { $_rename == "" } {
                return;
            }
            set label [$itk_component(legend) entry cget $_rename -label]
            foreach { l r w h } [$itk_component(legend) bbox $_rename] break
            set info(text) $label
            set info(x) [expr $l + [winfo rootx $itk_component(legend)]]
            set info(y) [expr $r + [winfo rooty $itk_component(legend)]] 
            set info(w) $w
            set info(h) $h
            return [array get info]
        }
        validate {
            if {[llength $args] != 1} {
                error "wrong # args: should be \"editor validate value\""
            }
        }
        apply {
            if {[llength $args] != 1} {
                error "wrong # args: should be \"editor apply value\""
            }
            set label [lindex $args 0]
            $itk_component(legend) entry configure $_rename -label $label
            set elem [$_tree label $_rename]
            $_graph element configure $elem -label $label
        }
        menu {
            eval tk_popup $itk_component(emenu) $args
        }
        default {
            error "bad option \"$option\": should be popup, activate, validate, apply, and menu"
        }
    }
}

#
# BuildPopup --
#
#
itcl::body Rappture::XyLegend::BuildPopup { popup } {
    set m $popup
    blt::combomenu $popup \
        -yscrollbar $m.ys \
        -xscrollbar $m.xs -height { 0 2.5i }
    blt::tk::scrollbar $m.xs
    blt::tk::scrollbar $m.ys
    $m add -text "Hide" -command [itcl::code $this Hide] 
    $m add -text "Show" -command [itcl::code $this Show] 
    $m add -text "Toggle" -command [itcl::code $this Toggle] 
    $m add -text "Unmap Hidden" -command [itcl::code $this UnmapHidden] \
        -variable [itcl::scope _unmapHidden] \
        -type checkbutton
    $m add -type separator 
    $m add -text "Lower" -command [itcl::code $this Lower] 
    $m add -text "Raise" -command [itcl::code $this Raise] 
    $m add -type separator 
    $m add -text "Average" -command [itcl::code $this Average]
    $m add -text "Difference" -command [itcl::code $this Difference]
    $m add -type separator 
    $m add -text "Recolor" -command [itcl::code $this Recolor]
    $m add -text "Rename" -command [itcl::code $this Rename]
    $m add -text "Delete" -command [itcl::code $this Delete]
    set _unmapHidden [$_graph cget -unmaphiddenelements]
    set cmd [itcl::code $this Toggle current]
    Check $m
}

#
# PopupMenu --
#
#	Builds the popup associated with the sensor map.  The menu
#	is first destroyed if one exists.  This is because the menu
#	bindings callbacks are specific to the current map.  Destroying 
#	the menu automatically removes the bindings and callbacks.
#
itcl::body Rappture::XyLegend::PopupMenu { x y } {
    set m .xylegendpopup
    if { [winfo exists $m]  } {
        destroy $m
    }
    BuildPopup $m
    blt::ComboMenu::popup $m $x $y
}
